"""Test the CLI can be run as a Python module."""

import subprocess
import sys


def test_can_run_as_python_module() -> None:
    """Run the CLI as a Python module."""
    result = subprocess.run(  # noqa: S603
        [sys.executable, "-m", "aiortm", "--help"],
        check=True,
        capture_output=True,
    )
    assert result.returncode == 0
    assert b"aiortm [OPTIONS]" in result.stdout



"""Provide a client for aiortm."""

from __future__ import annotations

import hashlib
from http import HTTPStatus
import json
import logging
from typing import Any, cast

from aiohttp import ClientResponse, ClientResponseError, ClientSession
from yarl import URL

from .exceptions import (
    APIAuthError,
    APIResponseError,
    TransportAuthError,
    TransportResponseError,
)
from .model import RTM
from .model.response import BaseErrorResponse

AUTH_URL = "https://www.rememberthemilk.com/services/auth/"
REST_URL = "https://api.rememberthemilk.com/services/rest/"
_LOGGER = logging.getLogger(__package__)


class Auth:
    """Represent the authentication manager."""

    def __init__(
        self,
        client_session: ClientSession,
        api_key: str,
        shared_secret: str,
        auth_token: str | None = None,
        permission: str = "delete",
    ) -> None:
        """Set up the auth manager."""
        self._client_session = client_session
        self.auth_token = auth_token
        self.permission = permission
        self.api_key = api_key
        self._shared_secret = shared_secret

    async def request(self, url: str, **kwargs: Any) -> ClientResponse:  # noqa: ANN401
        """Make a request."""
        headers: dict[str, Any] | None
        headers = {} if (headers := kwargs.get("headers")) is None else dict(headers)

        return await self._client_session.get(
            url,
            **kwargs,
            headers=headers,
        )

    async def authenticate_desktop(self) -> tuple[str, str]:
        """Authenticate as a desktop application."""
        data = await self.call_api("rtm.auth.getFrob", api_key=self.api_key)
        frob: str = data["frob"]
        url = self.generate_authorize_url(
            api_key=self.api_key,
            perms=self.permission,
            frob=frob,
        )
        return url, frob

    def generate_authorize_url(self, api_key: str, perms: str, frob: str) -> str:
        """Generate a URL for authorization."""
        params = {
            "api_key": api_key,
            "perms": perms,
            "frob": frob,
        }
        all_params = params | {"api_sig": self._sign_request(params)}
        return str(URL(AUTH_URL).with_query(all_params))

    async def get_token(self, frob: str) -> dict[str, Any]:
        """Fetch the authentication token with the frob."""
        data = await self.call_api("rtm.auth.getToken", api_key=self.api_key, frob=frob)
        auth_data: dict[str, Any] = data["auth"]
        self.auth_token = cast("str", auth_data["token"])
        return auth_data

    async def check_token(self) -> dict[str, Any]:
        """Check if auth token is valid.

        Will raise APIAuthError if the token is invalid.
        """
        data = await self.call_api_auth("rtm.auth.checkToken")
        auth_data: dict[str, Any] = data["auth"]
        return auth_data

    async def call_api_auth(self, api_method: str, **params: Any) -> dict[str, Any]:  # noqa: ANN401
        """Call an api method that requires authentication."""
        if self.auth_token is None:
            raise RuntimeError("Missing authentication token.")
        all_params = {"api_key": self.api_key, "auth_token": self.auth_token} | params
        return await self.call_api(api_method, **all_params)

    async def call_api(self, api_method: str, **params: Any) -> dict[str, Any]:  # noqa: ANN401
        """Call an api method."""
        # Remove empty values.
        params = {key: value for key, value in params.items() if value is not None}
        all_params = {"method": api_method} | params | {"format": "json"}
        all_params |= {"api_sig": self._sign_request(all_params)}
        response = await self.request(REST_URL, params=all_params)

        try:
            response.raise_for_status()
        except ClientResponseError as err:
            if err.status in (HTTPStatus.FORBIDDEN, HTTPStatus.UNAUTHORIZED):
                raise TransportAuthError(err) from err
            raise TransportResponseError(err) from err

        response_text = await response.text()

        if "rtm.auth" not in api_method:
            logged_response_text = response_text
            if self.api_key in response_text:
                logged_response_text = response_text.replace(self.api_key, "API_KEY")
            _LOGGER.debug("Response text: %s", logged_response_text)

        # API doesn't return a JSON encoded response.
        # It's text/javascript mimetype but with a JSON string in the text.
        data: dict[str, Any] = json.loads(response_text)["rsp"]

        if data["stat"] == "fail":
            error_response = BaseErrorResponse.from_dict(data)
            code = error_response.err.code
            if 98 <= code <= 100:
                raise APIAuthError(code, error_response.err.msg)
            raise APIResponseError(code, error_response.err.msg)
        return data

    def _sign_request(self, params: dict[str, Any]) -> str:
        """Return the string representing the signed request."""
        sorted_params = {key: params[key] for key in sorted(params)}
        param_string = "".join(
            f"{key}{val}" for key, val in sorted_params.items() if val is not None
        )
        data = f"{self._shared_secret}{param_string}".encode()
        return hashlib.md5(data).hexdigest()  # noqa: S324


class AioRTMClient:
    """Represent the aiortm client."""

    def __init__(
        self,
        auth: Auth,
    ) -> None:
        """Set up the client instance."""
        self.rtm = RTM(auth)



"""Provide a model for tasks."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any, Literal

from mashumaro import field_options
from mashumaro.mixins.json import DataClassJSONMixin

from .response import BaseResponse, TransactionResponse

if TYPE_CHECKING:
    from aiortm.client import Auth


@dataclass
class BaseModel(DataClassJSONMixin):
    """Represent a base model that turns empty string to None."""

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Turn empty string to None."""
        for k, v in d.items():
            if v == "":
                d[k] = None
            elif isinstance(v, list):
                d[k] = [cls.__pre_deserialize__(i) for i in v]
            elif isinstance(v, dict):
                d[k] = cls.__pre_deserialize__(v)
        return d


@dataclass
class TaskResponse(BaseModel):
    """Represent a response for a task."""

    id: int
    has_due_time: bool
    added: datetime = field(metadata={"deserialize": "ciso8601"})
    priority: Literal["N", "1", "2", "3"]
    postponed: bool
    due: datetime | None = field(metadata={"deserialize": "ciso8601"}, default=None)
    completed: datetime | None = field(
        metadata={"deserialize": "ciso8601"},
        default=None,
    )
    deleted: datetime | None = field(
        metadata={"deserialize": "ciso8601"},
        default=None,
    )
    estimate: str | None = None


@dataclass
class TaskSeriesResponse(BaseModel):
    """Represent a response for a task series."""

    id: int
    created: datetime = field(metadata={"deserialize": "ciso8601"})
    modified: datetime = field(metadata={"deserialize": "ciso8601"})
    name: str
    source: str
    tags: list[str]
    participants: list[str]
    notes: list[str]
    task: list[TaskResponse]
    location_id: str | None = None
    url: str | None = None


@dataclass
class TaskListResponse(BaseModel):
    """Represent a response for a task list."""

    id: int
    taskseries: list[TaskSeriesResponse]
    current: datetime | None = field(
        metadata={"deserialize": "ciso8601"},
        default=None,
    )


@dataclass
class RootTaskResponse(BaseModel):
    """Represent a response for the root tasks object."""

    rev: str
    task_list: list[TaskListResponse] = field(metadata=field_options(alias="list"))

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Ensure that taskseries exist."""
        task_list = d["list"]
        for item in task_list:
            if "taskseries" not in item:
                item["taskseries"] = []
        return d


@dataclass
class TaskModifiedResponse(BaseResponse):
    """Represent a response for a modified task."""

    transaction: TransactionResponse
    task_list: TaskListResponse = field(metadata=field_options(alias="list"))


@dataclass
class TasksResponse(BaseResponse):
    """Represent a response for a list of tasks."""

    tasks: RootTaskResponse


@dataclass
class Tasks:
    """Represent the tasks model."""

    api: Auth

    async def add(
        self,
        *,
        timeline: int,
        name: str,
        list_id: int | None = None,
        parse: bool | None = None,
    ) -> TaskModifiedResponse:
        """Add a task."""
        result = await self.api.call_api_auth(
            "rtm.tasks.add",
            timeline=timeline,
            name=name,
            list_id=list_id,
            parse="1" if parse else None,
        )
        return TaskModifiedResponse.from_dict(result)

    async def complete(
        self,
        *,
        timeline: int,
        list_id: int,
        taskseries_id: int,
        task_id: int,
    ) -> TaskModifiedResponse:
        """Complete a task."""
        result = await self.api.call_api_auth(
            "rtm.tasks.complete",
            timeline=timeline,
            list_id=list_id,
            taskseries_id=taskseries_id,
            task_id=task_id,
        )
        return TaskModifiedResponse.from_dict(result)

    async def delete(
        self,
        *,
        timeline: int,
        list_id: int,
        taskseries_id: int,
        task_id: int,
    ) -> TaskModifiedResponse:
        """Delete a task."""
        result = await self.api.call_api_auth(
            "rtm.tasks.delete",
            timeline=timeline,
            list_id=list_id,
            taskseries_id=taskseries_id,
            task_id=task_id,
        )
        return TaskModifiedResponse.from_dict(result)

    async def get_list(
        self,
        *,
        list_id: int | None = None,
        last_sync: datetime | None = None,
    ) -> TasksResponse:
        """Get a list of tasks."""
        last_sync_string: str | None = None
        if last_sync is not None:
            last_sync_string = last_sync.isoformat()

        result = await self.api.call_api_auth(
            "rtm.tasks.getList",
            list_id=list_id,
            last_sync=last_sync_string,
        )
        return TasksResponse.from_dict(result)

    async def set_name(
        self,
        *,
        timeline: int,
        list_id: int,
        taskseries_id: int,
        task_id: int,
        name: str,
    ) -> TaskModifiedResponse:
        """Rename a task."""
        result = await self.api.call_api_auth(
            "rtm.tasks.setName",
            timeline=timeline,
            list_id=list_id,
            taskseries_id=taskseries_id,
            task_id=task_id,
            name=name,
        )
        return TaskModifiedResponse.from_dict(result)



"""Test the client module."""

from __future__ import annotations

from collections.abc import Callable

from aioresponses import aioresponses
import pytest

from aiortm.client import AioRTMClient

from .util import load_fixture


@pytest.fixture(name="get_frob", scope="module")
def get_frob_fixture() -> str:
    """Return a response for rtm.auth.getFrob."""
    return load_fixture("auth/get_frob.json")


@pytest.fixture(name="get_token", scope="module")
def get_token_fixture() -> str:
    """Return a response for rtm.auth.getToken."""
    return load_fixture("auth/get_token.json")


@pytest.fixture(name="check_token", scope="module")
def check_token_fixture() -> str:
    """Return a response for rtm.auth.checkToken."""
    return load_fixture("auth/check_token.json")


async def test_authenticate_desktop(
    client: AioRTMClient,
    mock_response: aioresponses,
    get_frob: str,
    generate_url: Callable[..., str],
) -> None:
    """Test the authenticate desktop method."""
    mock_response.get(
        generate_url(
            api_key="test-api-key",
            method="rtm.auth.getFrob",
        ),
        body=get_frob,
    )

    url, frob = await client.rtm.api.authenticate_desktop()

    assert url == (
        "https://www.rememberthemilk.com/services/auth/?"
        "api_key=test-api-key&perms=delete&frob=test-frob"
        "&api_sig=9d28d5a58ff2efc3fed277a58c3ce818"
    )
    assert frob == "test-frob"


async def test_get_token(
    client: AioRTMClient,
    mock_response: aioresponses,
    get_token: str,
    generate_url: Callable[..., str],
) -> None:
    """Test the get_token method."""
    mock_response.get(
        generate_url(
            api_key="test-api-key",
            frob="test-frob",
            method="rtm.auth.getToken",
        ),
        body=get_token,
    )

    result = await client.rtm.api.get_token("test-frob")

    assert result["token"] == "test-token"
    assert result["perms"] == "delete"
    assert result["user"]["id"] == "1234567"
    assert result["user"]["username"] == "johnsmith"
    assert result["user"]["fullname"] == "John Smith"


async def test_check_token(
    client: AioRTMClient,
    mock_response: aioresponses,
    check_token: str,
    generate_url: Callable[..., str],
) -> None:
    """Test the check_token method."""
    mock_response.get(
        generate_url(
            api_key="test-api-key",
            auth_token="test-token",
            method="rtm.auth.checkToken",
        ),
        body=check_token,
    )

    result = await client.rtm.api.check_token()

    assert result["token"] == "test-token"
    assert result["perms"] == "delete"
    assert result["user"]["id"] == "1234567"
    assert result["user"]["username"] == "johnsmith"
    assert result["user"]["fullname"] == "John Smith"



"""Test the CLI."""

from typer.testing import CliRunner

from aiortm.cli import cli

runner = CliRunner()


def test_help() -> None:
    """The help message includes the CLI name."""
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert "Authorize the app." in result.stdout



"""Provide a model for contacts."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from mashumaro.mixins.json import DataClassJSONMixin

from .response import BaseResponse, TransactionResponse

if TYPE_CHECKING:
    from aiortm.client import Auth


@dataclass
class ContactResponse(DataClassJSONMixin):
    """Represent a response for a contact."""

    id: int
    fullname: str
    username: str


@dataclass
class ContactAddResponse(BaseResponse):
    """Represent a response for add contact."""

    transaction: TransactionResponse
    contact: ContactResponse


@dataclass
class ContactsResponse(BaseResponse):
    """Represent a response for a list of contacts."""

    # The API inconsistently returns an empty list when no contacts are present.
    contacts: list[ContactResponse]

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Ensure that contacts is a list of contacts."""
        contacts = d["contacts"]
        if isinstance(contacts, dict):
            d["contacts"] = contacts["contact"]
        return d


@dataclass
class ContactDeleteResponse(BaseResponse):
    """Represent a response for delete contact."""

    transaction: TransactionResponse


@dataclass
class Contacts:
    """Represent the contacts model."""

    api: Auth

    async def add(self, *, timeline: int, contact: str) -> ContactAddResponse:
        """Add a contact."""
        result = await self.api.call_api_auth(
            "rtm.contacts.add",
            timeline=timeline,
            contact=contact,
        )
        return ContactAddResponse.from_dict(result)

    async def delete(self, *, timeline: int, contact_id: int) -> ContactDeleteResponse:
        """Delete a contact."""
        result = await self.api.call_api_auth(
            "rtm.contacts.delete",
            timeline=timeline,
            contact_id=contact_id,
        )
        return ContactDeleteResponse.from_dict(result)

    async def get_list(self) -> ContactsResponse:
        """Get a list of contacts."""
        result = await self.api.call_api_auth("rtm.contacts.getList")
        return ContactsResponse.from_dict(result)
